import logging

logger = logging.getLogger(__name__)

try:
    import quadpy as qp
    import numpy
    class QPQuadPoint:
        def __init__(self,p,w):
            self.p_ = p
            self.w_ = w
        @property
        def position(self):
            return self.p_
        @property
        def weight(self):
            return self.w_
    _cache = {}
    class QPQuadPyQuadrature:
        def __init__(self,quad,order,method,vertices,transform):
            self.quad_ = quad
            self.order_ = order
            self.method_ = method
            self.vertices_ = numpy.array(vertices)
            # here an error will occur if the method is invalid - how to catch?
            self.quadrature_ = getattr(quad,method)
            try:
                self.points_ = transform.transform(self.quadrature_.points.T, self.vertices_)
            except ValueError:
                self.points_ = transform.transform(self.quadrature_.points.T, self.vertices_.T).T
            try:
                self.weights_ = transform.get_detJ(self.quadrature_.points.T, self.vertices_)*self.quadrature_.weights
            except AttributeError:
                self.weights_ = transform.get_vol(self.vertices_)*self.quadrature_.weights
            self.quadPoints_ = [ QPQuadPoint(p,w) for p,w in zip(self.points_,self.weights_) ]
            self.points_ = self.points_.transpose().copy()
        def get(self):
            return self.points_, self.weights_
        def apply(self,entity,f):
            ie = entity.geometry.integrationElement
            f_e = lambda x: f(entity,x)*ie(x)
            return self.quad_.integrate(f_e,self.vertices_,self.quadrature_)
        def __iter__(self):
            return self.quadPoints_.__iter__()
        @property
        def order(self):
            return self.quadrature_.degree
        def name(self):
            return self.method_
    def rule(gt,quadDescription):
        try:
            gt = gt.type
        except AttributeError:
            pass
        try:
            quad = _cache[(gt,quadDescription)]
            return quad
        except KeyError:
            pass
        order, method = quadDescription
        dim = gt.dim
        if gt.isLine:
            vertices = [0,1]
            quad = qp.line_segment
            from quadpy.ncube import transform
        elif gt.isTriangle:
            vertices = [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
            quad = qp.triangle
            from quadpy.nsimplex import transform
        elif gt.isQuadrilateral:
            vertices = qp.quadrilateral.rectangle_points([0.0, 1.0], [0.0, 1.0])
            quad = qp.quadrilateral
            from quadpy.ncube import transform
        elif gt.isTetrahedron:
            vertices = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
            quad = qp.tetrahedron
            from quadpy.nsimplex import transform
        elif gt.isHexahedron:
            vertices = qp.hexahedron.cube_points([0.0, 1.0], [0.0, 1.0], [0.0, 1.0])
            quad = qp.hexahedron
            from quadpy.ncube import transform
        elif gt.isPyramid:
            vertices = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
            quad = qp.pyramid
            raise ValueError("prism quadratures not yet fully supported")
        elif gt.isPrism:
            vertices = [
                [0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0], [1.0, 0.0, 1.0], [0.0, 1.0, 1.0]
                       ]
            quad = qp.wedge
            raise ValueError("prism quadratures not yet fully supported")
        else:
            raise ValueError("no quadpy quadrature available for the geometryType " + str(gt))
        if not method and not order:
            return quad
        ret = QPQuadPyQuadrature(quad,order,method,vertices,transform)
        _cache[(gt,quadDescription)] = ret
        return ret

    def rules(methods):
        def r(entity):
            try:
                return rule(entity.type, methods[entity.type])
            except AttributeError:
                return rule(entity, methods[entity])
        return r

except ImportError as e:
    logger.warning('Unable to import quadpy: ' + " ".join(str(e).splitlines()))
    raise ImportError("Unable to import quadpy module")



from dune.geometry import referenceElement
from dune.geometry import vertex, line, triangle, quadrilateral, tetrahedron, pyramid, prism, hexahedron, none

def test(r):
    for codim in range(r.dimension+1):
        types = r.types(codim)
        if len(types) != r.size(codim):
            raise Exception("types tuple has wrong size")
        for i in range(len(types)):
            if types[i] != r.type(i, codim):
                raise Exception("types tuple has wrong content")

    for codim in range(r.dimension+1):
        positions = r.positions(codim)
        if len(positions) != r.size(codim):
            raise Exception("positions tuple has wrong size")
        for i in range(len(positions)):
            if positions[i] != r.position(i, codim):
                raise Exception("positions tuple has wrong content")

    if r.dimension > 0:
        normals = r.integrationOuterNormals
        if len(normals) != r.size(1):
            raise Exception("integrationOuterNormals has wrong size")
        for i in range(len(normals)):
            if normals[i] != r.integrationOuterNormal(i):
                raise Exception("integrationOuterNormals has wrong content")

test(referenceElement(vertex))
test(referenceElement(line))
test(referenceElement(triangle))
test(referenceElement(quadrilateral))
test(referenceElement(tetrahedron))
test(referenceElement(pyramid))
test(referenceElement(prism))
test(referenceElement(hexahedron))

for dim in range(4):
    if referenceElement(none(dim)) is not None:
        raise Exception("got reference element for geometry type none")



from ..generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
def module(dim):
    typeName = "Dune::Geo::ReferenceElement<Dune::Geo::ReferenceElementImplementation<double," + str(dim) + "> >"
    includes = ["dune/python/geometry/referenceelements.hh"]
    typeHash = "referenceelements_" + hashIt(typeName)
    generator = SimpleGenerator("ReferenceElements", "Dune::Python")
    m = generator.load(includes, typeName, typeHash)
    return m

_duneReferenceElements = {}
def referenceElement(geometryType):
    try:
        geometryType = geometryType.type
    except:
        pass
    try:
        ref = _duneReferenceElements[geometryType]
    except KeyError:
        ref = module(geometryType.dim).general(geometryType)
        _duneReferenceElements[geometryType] = ref
    return ref



from dune.geometry import *

# make sure t can be reconstructed from str(t)
for t in (vertex, line, triangle, quadrilateral, tetrahedron, pyramid, prism, hexahedron):
    assert GeometryType(str(t)) == t

for d in range(10):
    assert GeometryType(str(simplex(d))) == simplex(d)
    assert GeometryType(str(cube(d))) == cube(d)
    assert GeometryType(str(none(d))) == none(d)

# make sure simplices with special names can be constructed by general mechanism
for d, t in enumerate((vertex, line, triangle, tetrahedron)):
    assert GeometryType("simplex(" + str(d) + ")") == t
    assert GeometryType("general( 0, " + str(d) + ")") == t

# make sure cube with special names can be constructed by general mechanism
for d, t in enumerate((vertex, line, quadrilateral, hexahedron)):
    assert GeometryType("cube(" + str(d) + ")") == t
    assert GeometryType("general( " + str(2**d-1) + ", " + str(d) + ")") == t



import time, math, numpy
import dune.geometry as geo

def monomial(p):
    def function(point):
        return sum( x**p for x in point)
    return function

result = {3:   # integral for sum_i x_i^p over reference element
           {geo.line: 1./4.,
            geo.triangle: 0.1,
            geo.quadrilateral: 1./2.,
            geo.tetrahedron: 1./40.,
            geo.pyramid: None,
            geo.prism: None,
            geo.hexahedron: 3./4.,
           },
          4:
           {geo.line: 1./5.,
            geo.triangle: 1./15.,
            geo.quadrilateral: 2./5.,
            geo.tetrahedron: 1./70.,
            geo.pyramid: None,
            geo.prism: None,
            geo.hexahedron: 3./5.,
           },
         }



for order in [3,4]:
    rules = geo.quadratureRules(order)
    p = monomial(order)
    for t in (geo.line, geo.triangle, geo.quadrilateral,
              geo.tetrahedron, geo.pyramid, geo.prism, geo.hexahedron):
        value1 = 0
        for q in rules(t):
            value1 += p(q.position)*q.weight
        hatxs, hatws = rules(t).get()
        value2 = numpy.sum(p(hatxs) * hatws, axis=-1)
        # print(order,t,value2)
        assert abs(value1-value2)<1e-14
        if result[order][t] is not None:
            assert abs(result[order][t] - value1)<1e-12
